// Group.cpp: implementation of the CGroup class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Group.h"
#include "utils.h"
#include "pidl.h"
#include "lsapiex.h"
#include "main.h"
#include "DropTarget.h"

#define REGKEY_MODULE	"SOFTWARE\\LiteStep\\Clickonic\\"
#ifndef LVS_EX_SNAPTOGRID
#define LVS_EX_SNAPTOGRID       0x00080000  // Icons automatically snap to grid.
#endif

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Constructor/destructor

CGroup::CGroup(const char *szName)
{
	IShellBrowserImpl();
	m_bInitialized = false;
	m_pDropTarget = NULL;
	m_nLastItem = -1;

	wpOrigListView_Proc = NULL;

	utils::CopyString(&m_szName, szName);
	
	if (!ReadSettings())
	{
		return;
	}

	m_hWnd = utils::GetLitestepDesktop();

	if (!InitFolderView())
	{
		return;
	}

	m_bInitialized = true;
}


CGroup::~CGroup()
{
	SaveState();

	if ( wpOrigListView_Proc != NULL )
	{
		// Remove the subclass from the listview. 
        SetWindowLong(m_hwndListView, GWL_WNDPROC, (LONG) wpOrigListView_Proc); 
	}

	if (m_pDropTarget)
	{
		m_pDropTarget->Revoke(m_hwndView);
		delete m_pDropTarget;
	}
	if (m_pView2)
	{
		m_pView2->Release();
	}
	if (m_pView)
	{
		m_pView->DestroyViewWindow();
		m_pView->Release();
	}
	if (m_pFolder)
	{
		m_pFolder->Release();
	}
	if (m_hFont)
	{
		DeleteObject(m_hFont);
	}
	if (m_szName)
	{
		delete m_szName;
	}
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Settings parser...

bool CGroup::ReadSettings()
{
	char szBuf[MAX_LINE_LENGTH], szExpanded[MAX_LINE_LENGTH];
	LiteStep::GetPrefixedRCString(szBuf, m_szName, "Folder", ".desktop");
	VarExpansionEx(szExpanded, szBuf, MAX_LINE_LENGTH);

	if (_stricmp(szExpanded, ".desktop") == 0)
	{
		SHGetSpecialFolderPath(NULL, m_szFolderLocation, CSIDL_DESKTOP, 0);
		SHGetSpecialFolderLocation(NULL, CSIDL_DESKTOP, &m_pidlFolder);
	}
	else
	{
		strcpy_s(m_szFolderLocation, szExpanded);
		m_pidlFolder = PIDL::ParseDisplayName(szExpanded);
	}
	if (!m_pidlFolder)
	{
		return false;
	}

	m_nX		= LiteStep::GetPrefixedRCCoordinate(m_szName, "X", 0, SCREEN_WIDTH);
	m_nY		= LiteStep::GetPrefixedRCCoordinate(m_szName, "Y", 0, SCREEN_HEIGHT);

	m_nWidth	= LiteStep::GetPrefixedRCInt(m_szName, "Width",  SCREEN_WIDTH);
	m_nHeight	= LiteStep::GetPrefixedRCInt(m_szName, "Height", SCREEN_HEIGHT);

	m_nSpacingX	= LiteStep::GetPrefixedRCInt(m_szName, "SpacingX",  -1);
	m_nSpacingY	= LiteStep::GetPrefixedRCInt(m_szName, "SpacingY", -1);
	
	m_bHidden	= LiteStep::GetPrefixedRCBool(m_szName, "StartHidden", FALSE);
	m_bNoScroll	= LiteStep::GetPrefixedRCBool(m_szName, "NoScroll", FALSE);

	m_bUseDelete	= LiteStep::GetPrefixedRCBool(m_szName, "UseDelete", TRUE);
	m_bUseRename	= LiteStep::GetPrefixedRCBool(m_szName, "UseRename", TRUE);

	/*
	m_bUseRegistryIcons		= LiteStep::GetPrefixedRCBool(m_szName, "UseRegistryIcons", FALSE);

	if (!m_bUseRegistryIcons) {
		m_bHideRecycleBin		= LiteStep::GetPrefixedRCBool(m_szName, "HideRecycleBin", FALSE);
		m_bHideMyDocuments		= LiteStep::GetPrefixedRCBool(m_szName, "HideMyDocuments", FALSE);
		m_bHideInternetExplorer	= LiteStep::GetPrefixedRCBool(m_szName, "HideInternetExplorer", FALSE);
		m_bHideMyNetworkPlaces	= LiteStep::GetPrefixedRCBool(m_szName, "HideMyNetworkPlaces", FALSE);
		m_bHideMyComputer		= LiteStep::GetPrefixedRCBool(m_szName, "HideMyComputer", FALSE);
	}
	else {
		ReadIconSettingsFromRegistry();
	}
	*/

	m_bDontUseRecycleBin	= LiteStep::GetPrefixedRCBool(m_szName, "DontUseRecycleBin", FALSE);
	m_bConfirmFileDelete	= LiteStep::GetPrefixedRCBool(m_szName, "ConfirmFileDelete", FALSE);
	m_bSimpleProgress		= LiteStep::GetPrefixedRCBool(m_szName, "SimpleProgress", FALSE);
	m_bNoConnectedFiles		= LiteStep::GetPrefixedRCBool(m_szName, "NoConnectedFiles", FALSE);
	m_bVirtualFileDelete	= LiteStep::GetPrefixedRCBool(m_szName, "VirtualFileDelete", FALSE);

	m_bSnapToGrid		= LiteStep::GetPrefixedRCBool(m_szName, "SnapToGrid", TRUE);
	m_bHotTracking		= LiteStep::GetPrefixedRCBool(m_szName, "HotTracking", FALSE);
	m_nHotTrackingTime	= LiteStep::GetPrefixedRCInt(m_szName, "HotTrackingTime", -1);

	m_bExplorerDesktop = LiteStep::GetPrefixedRCBool(m_szName, "ExplorerDesktop", FALSE);

	LiteStep::GetPrefixedRCString(szBuf, m_szName, "Background", "");
	utils::CopyString(&m_szBackground, szBuf);
	if (m_szBackground && m_szBackground[0])
	{
		m_bTiled = LiteStep::GetPrefixedRCBool(m_szName, "Tiled", false);
		m_nOffsetPercentX = LiteStep::GetPrefixedRCInt(m_szName, "OffsetPercentX", 0);
		m_nOffsetPercentY = LiteStep::GetPrefixedRCInt(m_szName, "OffsetPercentY", 0);
	}
	m_crBackgroundColor= LiteStep::GetPrefixedRCColor(m_szName, "BackgroundColor", CLR_NONE);
	
	// Get the font
	LiteStep::GetPrefixedRCString(szBuf, m_szName, "Font", "Arial");
	m_crTextColor= LiteStep::GetPrefixedRCColor(m_szName, "FontColor", RGB(255, 255, 255));
	m_crTextBackgroundColor= LiteStep::GetPrefixedRCColor(m_szName, "FontBackgroundColor", CLR_NONE);
	int nFontSize		= LiteStep::GetPrefixedRCInt(m_szName, "FontSize", 14);
	bool bFontItalic	= LiteStep::GetPrefixedRCBool(m_szName, "FontItalic", false);
	bool bFontBold		= LiteStep::GetPrefixedRCBool(m_szName, "FontBold", false);
	
	m_dListStyleFlags = 0;
	if (m_bHotTracking)
		m_dListStyleFlags |= LVS_EX_TRACKSELECT;
	if (m_bSnapToGrid)
		m_dListStyleFlags |= LVS_EX_SNAPTOGRID;

	m_hFont = CreateFont(
		nFontSize,
		0, 0, 0,
		bFontBold ? FW_BOLD : FW_NORMAL,
		bFontItalic ? TRUE: FALSE,
		FALSE, FALSE,
		DEFAULT_CHARSET,
		OUT_DEFAULT_PRECIS,
		CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY,
		DEFAULT_PITCH | FF_DONTCARE,
		szBuf);
	if (!m_hFont)
	{
		utils::ErrorBoxPrint("Could not load font %s", szBuf);
	}

	return true;
}

void CGroup::ReadIconSettingsFromRegistry() {
	/*bool szBuffer[1];
	DWORD dwSize, dwType;
	dwSize = sizeof(szBuffer);
	SHGetValue(HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\HideDesktopIcons\\NewStartPanel",
		"{871C5380-42A0-1069-A2EA-08002B30309D}", &dwType, szBuffer, &dwSize);
	m_bHideInternetExplorer = szBuffer[0];
	SHGetValue(HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\HideDesktopIcons\\NewStartPanel",
		"{20D04FE0-3AEA-1069-A2D8-08002B30309D}", &dwType, szBuffer, &dwSize);
	m_bHideMyComputer = szBuffer[0];
	SHGetValue(HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\HideDesktopIcons\\NewStartPanel",
		"{450D8FBA-AD25-11D0-98A8-0800361B1103}", &dwType, szBuffer, &dwSize);
	m_bHideMyDocuments = szBuffer[0];
	SHGetValue(HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\HideDesktopIcons\\NewStartPanel",
		"{208D2C60-3AEA-1069-A2D7-08002B30309D}", &dwType, szBuffer, &dwSize);
	m_bHideMyNetworkPlaces = szBuffer[0];
	SHGetValue(HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\HideDesktopIcons\\NewStartPanel",
		"{645FF040-5081-101B-9F08-00AA002F954E}", &dwType, szBuffer, &dwSize);
	m_bHideRecycleBin = szBuffer[0];
	*/
	/*
	// This is better... really... I just can't get it working :/
	HKEY hKey;
	char* szBuffer;
	DWORD dwSize;
	VALENT veList[4];

	RegOpenKeyEx(
		HKEY_CURRENT_USER,
		"Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\HideDesktopIcons\\NewStartPanel",
		0,
		KEY_ALL_ACCESS,
		&hKey);

	veList[0].ve_valuename = "{20D04FE0-3AEA-1069-A2D8-08002B30309D}";
	veList[0].ve_type = REG_DWORD;
	veList[1].ve_valuename = "{871C5380-42A0-1069-A2EA-08002B30309D}";
	veList[1].ve_type = REG_DWORD;
	veList[2].ve_valuename = "{450D8FBA-AD25-11D0-98A8-0800361B1103}";
	veList[2].ve_type = REG_DWORD;
	veList[3].ve_valuename = "{645FF040-5081-101B-9F08-00AA002F954E}";
	veList[3].ve_type = REG_DWORD;
	veList[4].ve_valuename = "{208D2C60-3AEA-1069-A2D7-08002B30309D}";
	veList[4].ve_type = REG_DWORD;

	RegQueryMultipleValues(hKey, &veList, 5, NULL, &dwSize);

	szBuffer = reinterpret_cast<char*>(HeapAlloc(GetProcessHeap(),0, dwSize));

	RegQueryMultipleValues(hKey, &veList, 5, szBuffer, &dwSize);

	//char test[256];
	//sprintf_s(test, "Testing: %s", veList[0].ve_valueptr);

	MessageBox(0, szBuffer, "", 0);
	*/
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Sorting routines...

int SortPIDLs(LPSHELLFOLDER psf, LPARAM lColumn, LPITEMIDLIST pidl1, LPITEMIDLIST pidl2)
{
	HRESULT hres = psf->CompareIDs(lColumn, pidl1, pidl2);
	//utils::log("%d :: (%d %d) -> %d::%d -> %d", psf, pidl1, pidl2, HRESULT_CODE(hres), static_cast<signed short>(HRESULT_CODE(hres)), hres);
	return static_cast<signed short>(HRESULT_CODE(hres));
}

int CALLBACK SortByNameCompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	return SortPIDLs(
		reinterpret_cast<LPSHELLFOLDER>(lParamSort), 
		0,
		reinterpret_cast<LPITEMIDLIST>(lParam1),
		reinterpret_cast<LPITEMIDLIST>(lParam2)
	);
}

int CALLBACK SortBySizeCompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	return SortPIDLs(
		reinterpret_cast<LPSHELLFOLDER>(lParamSort), 
		1,
		reinterpret_cast<LPITEMIDLIST>(lParam1),
		reinterpret_cast<LPITEMIDLIST>(lParam2)
	);
}

int CALLBACK SortByTypeCompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	return SortPIDLs(
		reinterpret_cast<LPSHELLFOLDER>(lParamSort), 
		2,
		reinterpret_cast<LPITEMIDLIST>(lParam1),
		reinterpret_cast<LPITEMIDLIST>(lParam2)
	);
}

int CALLBACK SortByDateCompareFunc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	return SortPIDLs(
		reinterpret_cast<LPSHELLFOLDER>(lParamSort), 
		3,
		reinterpret_cast<LPITEMIDLIST>(lParam1),
		reinterpret_cast<LPITEMIDLIST>(lParam2)
	);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Bangs

void CGroup::Show()
{
	ShowWindow(m_hwndView, SW_SHOW);
	m_bHidden = false;
}

void CGroup::Hide()
{
	ShowWindow(m_hwndView, SW_HIDE);
	m_bHidden = true;
}


void CGroup::Toggle()
{
	ShowWindow(m_hwndView, m_bHidden ? SW_SHOW : SW_HIDE);
	m_bHidden = !m_bHidden;
}

/*void CGroup::Hook(HWND hParent)
{
	if (hParent != GetParent(m_hwndView))
	{
		SetWindowLong(m_hwndView, GWL_STYLE, (GetWindowLong(m_hwndView, GWL_STYLE) & ~WS_POPUP)|WS_CHILD);
		SetParent(m_hwndView, hParent);
	}
}*/

void CGroup::ToggleExplorerDesktop()
{
	m_bExplorerDesktop = !m_bExplorerDesktop;
}

void CGroup::ExplorerDesktop( bool bEnabled ) {
	m_bExplorerDesktop = bEnabled;
}

void CGroup::HotTracking( bool bEnabled ) {
	if (bEnabled) {
		if (!m_bHotTracking) {
			m_dListStyleFlags |= LVS_EX_TRACKSELECT;
			m_bHotTracking = true;
		}
	}
	else {
		if (m_bHotTracking) {
			m_dListStyleFlags ^= LVS_EX_TRACKSELECT;
			m_bHotTracking = false;
		}
	}
	ListView_SetExtendedListViewStyle(m_hwndListView, m_dListStyleFlags);
}

void CGroup::HotTrackingTime( int iTime ) {
	m_nHotTrackingTime = iTime;
	ListView_SetHoverTime(m_hwndListView, m_nHotTrackingTime);
}

void CGroup::ConfirmFileDelete( bool bEnabled ) {
	m_bConfirmFileDelete = bEnabled;
}

void CGroup::DontUseReycleBin( bool bEnabled ) {
	m_bDontUseRecycleBin = bEnabled;
}

void CGroup::NoConnectedFiles( bool bEnabled ) {
	m_bNoConnectedFiles = bEnabled;
}

void CGroup::SimpleProgress( bool bEnabled ) {
	m_bSimpleProgress = bEnabled;
}

void CGroup::VirtualFileDelete ( bool bEnabled ) {
	m_bVirtualFileDelete = bEnabled;
}

void CGroup::SnapToGrid( bool bEnabled ) {
	if (bEnabled) {
		if (!m_bSnapToGrid) {
			m_dListStyleFlags |= LVS_EX_SNAPTOGRID;
			m_bSnapToGrid = true;
		}
	}
	else {
		if (m_bSnapToGrid) {
			m_dListStyleFlags ^= LVS_EX_SNAPTOGRID;
			m_bSnapToGrid = false;
		}
	}
	ListView_SetExtendedListViewStyle(m_hwndListView, m_dListStyleFlags);
}
void CGroup::UseDelete (bool bEnabled) {
	m_bUseDelete = bEnabled;
}
void CGroup::UseRename (bool bEnabled) {
	m_bUseRename = bEnabled;
}

void CGroup::Reset()
{
	char szKeyName[MAX_PATH];
	StringCchPrintf(szKeyName, MAX_PATH, "%s%s", REGKEY_MODULE, m_szName);
	// delete data stored in the registry
	RegDeleteKey(HKEY_CURRENT_USER, szKeyName);
	// sort'em by name
	ListView_SortItems(m_hwndListView, SortByNameCompareFunc, m_pFolder);
}

void CGroup::DeleteFiles() {
	char szFileName[MAX_PATH];
	char szFullCommand[MAX_LINE_LENGTH] = {0};
	int iItem = -1;
	UINT i, j;
	UINT iNumSelected = ListView_GetSelectedCount(m_hwndListView);
	UINT iCommandPos = 0;
	UINT iFakeFiles[256], iRealFiles[256];
	UINT iFakeFileCounter = 0, iRealFileCounter = 0;
	SHFILEOPSTRUCT SHFileDelete;
	STRRET strDispName;
	LVITEM lvitem;
	size_t FileNameSize = 0;

	if (iNumSelected != 0) {
		if (m_bVirtualFileDelete) {
			DeleteFilesVirtually();
		}
		else {
			for (i = 0; i < iNumSelected; i++) {
				iItem = ListView_GetNextItem(m_hwndListView, iItem, LVNI_SELECTED);
				ZeroMemory(&lvitem, sizeof(LVITEM));
				lvitem.mask = LVIF_PARAM;
				lvitem.iItem = iItem;
				lvitem.iSubItem = 0;
				ListView_GetItem(m_hwndListView, &lvitem);
				m_pFolder->GetDisplayNameOf((LPCITEMIDLIST)(lvitem.lParam), SHGDN_FORPARSING, &strDispName);
				StrRetToBuf(&strDispName, (LPCITEMIDLIST)(lvitem.lParam), szFileName, MAX_PATH);
				StringCchLength(szFileName, MAX_PATH, &FileNameSize);
				if (utils::File_Exists(szFileName)) {
					iRealFiles[iRealFileCounter] = iItem;
					iRealFileCounter++;
					for(j = 0; j <= FileNameSize; j++) {
						szFullCommand[iCommandPos++] = szFileName[j];
					}
				}
				else {
					// This has got to go!, but it will serve as a temporary fix until i figure out how to handle
					// those ::{} files.
					iFakeFiles[iFakeFileCounter] = iItem;
					iFakeFileCounter++;
				}
			}
			for (i = 0; i < iFakeFileCounter; i++) {
				for (j = 0; j < iRealFileCounter; j++) {
					if (iRealFiles[j] < iFakeFiles[i])
						iFakeFiles[i]--;
				}
			}
			if (iRealFileCounter != 0) {
				ZeroMemory(&SHFileDelete, sizeof(SHFILEOPSTRUCT));
				SHFileDelete.wFunc = FO_DELETE;
				SHFileDelete.hwnd = NULL;
				SHFileDelete.pFrom = szFullCommand;
				SHFileDelete.pTo = NULL;
				SHFileDelete.fFlags = 0;
				if (m_bSimpleProgress)
					SHFileDelete.fFlags |= FOF_SIMPLEPROGRESS;
				if (!m_bDontUseRecycleBin)
					SHFileDelete.fFlags |= FOF_ALLOWUNDO;
				if (!m_bConfirmFileDelete)
					SHFileDelete.fFlags |= FOF_NOCONFIRMATION | FOF_WANTNUKEWARNING;
				if (m_bNoConnectedFiles)
					SHFileDelete.fFlags |= FOF_NO_CONNECTED_ELEMENTS;
				if (SHFileOperation(&SHFileDelete) == 0) {
					if (!SHFileDelete.fAnyOperationsAborted) {
						for (i = 0; i < iFakeFileCounter; i++) {
							ListView_DeleteItem(m_hwndListView, iFakeFiles[i]);
						}
					}
				}
			}
			else {
				DeleteFilesVirtually();
			}
		}
	}
}
void CGroup::DeleteFilesVirtually() {
	char szFilename[256], szTemp[512];
	int nSelected = ListView_GetSelectedCount(m_hwndListView), iItem = -1, i;
	if (nSelected == 1) {
		ListView_GetItemText(m_hwndListView, ListView_GetNextItem(m_hwndListView, iItem, LVNI_SELECTED), 0, szFilename, 256);
		sprintf_s(szTemp, "Are you sure that you want to delete %s?", szFilename);
	}
	else {
		sprintf_s(szTemp, "Are you sure that you want to delete these %i files?", nSelected);
	}
	if (MessageBox(NULL, szTemp, "Confirm File Delete", MB_YESNO | MB_ICONQUESTION) == IDYES) {	
		for (i = 0; i < nSelected; i++) {
			iItem = ListView_GetNextItem(m_hwndListView, iItem, LVNI_SELECTED);
			ListView_DeleteItem(m_hwndListView, iItem);
		}
	}
}
/*void CGroup::RemoveUnwantedFiles() {

	STRRET strDispName;
	char szFileName[MAX_PATH];
	char szMyDocumentsPath[MAX_PATH];
	size_t FileNameSize;
	LVITEM lvitem;
	int nCount = ListView_GetItemCount(m_hwndListView);
	bool bDelete;

	SHGetSpecialFolderPath(NULL, szMyDocumentsPath, CSIDL_PERSONAL, false);

	for (int i = 0; i < nCount; i++) {
		ZeroMemory(&lvitem, sizeof(LVITEM));
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = i;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);
		m_pFolder->GetDisplayNameOf((LPCITEMIDLIST)(lvitem.lParam), SHGDN_FORPARSING, &strDispName);
		StrRetToBuf(&strDispName, (LPCITEMIDLIST)(lvitem.lParam), szFileName, MAX_PATH);
		StringCchLength(szFileName, MAX_PATH, &FileNameSize);

		bDelete = false;

		if (m_bHideMyDocuments && !strcmp(szFileName, szMyDocumentsPath)) {
			bDelete = true;
		}
		else if (m_bHideInternetExplorer && !strcmp(szFileName, "::{871C5380-42A0-1069-A2EA-08002B30309D}")) {
			bDelete = true;
		}
		else if (m_bHideMyNetworkPlaces && !strcmp(szFileName, "::{208D2C60-3AEA-1069-A2D7-08002B30309D}")) {
			bDelete = true;
		}
		else if (m_bHideMyComputer && !strcmp(szFileName, "::{20D04FE0-3AEA-1069-A2D8-08002B30309D}")) {
			bDelete = true;
		}
		else if (m_bHideRecycleBin && !strcmp(szFileName, "::{645FF040-5081-101B-9F08-00AA002F954E}")) {
			bDelete = true;
		}
		if (bDelete) {
			//MessageBox(NULL, szFileName, "Removing this file", MB_OK);
			// It's a bit of a hack... but it's better than nothing
			ListView_DeleteItem(m_hwndListView, i);
			nCount = ListView_GetItemCount(m_hwndListView);
			i--;
		}
	}
}*/
void CGroup::EchoFile() {
	char szFileName[MAX_PATH];
	int iItem = -1;
	STRRET strDispName;
	LVITEM lvitem;
	size_t FileNameSize = 0;
	SFGAOF flags;

	iItem = ListView_GetNextItem(m_hwndListView, iItem, LVNI_SELECTED);
	ZeroMemory(&lvitem, sizeof(LVITEM));
	lvitem.mask = LVIF_PARAM;
	lvitem.iItem = iItem;
	lvitem.iSubItem = 0;
	ListView_GetItem(m_hwndListView, &lvitem);
	m_pFolder->GetDisplayNameOf((LPCITEMIDLIST)(lvitem.lParam), SHGDN_FORPARSING, &strDispName);
	flags = SFGAO_FILESYSTEM;
	m_pFolder->GetAttributesOf(1, (LPCITEMIDLIST*)(&lvitem.lParam), &flags);
	StrRetToBuf(&strDispName, (LPCITEMIDLIST)(lvitem.lParam), szFileName, MAX_PATH);
	StringCchLength(szFileName, MAX_PATH, &FileNameSize);

	
	//if (!(flags & SFGAO_FILESYSTEM))
	//	MessageBox(NULL, "It's not a filesystem object!!!", "Debug - Info", MB_OK);

	MessageBox(NULL, szFileName, "Debug - Filename", MB_OK);
}

void CGroup::CopyFiles(bool cut)
{
	char szFileName[MAX_PATH];
	char szFullCommand[MAX_LINE_LENGTH];
	int iItem = -1;
	UINT i, j;
	UINT iNumSelected = ListView_GetSelectedCount(m_hwndListView);
	UINT iCommandPos = 0;
	UINT iFiles[256], iFileCounter = 0;
	STRRET strDispName;
	LVITEM lvitem;
	size_t FileNameSize = 0;

	if (!(iNumSelected > 0))
		return;

	for (i = 0; i < iNumSelected; i++) {
		iItem = ListView_GetNextItem(m_hwndListView, iItem, LVNI_SELECTED);
		ZeroMemory(&lvitem, sizeof(LVITEM));
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = iItem;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);
		m_pFolder->GetDisplayNameOf((LPCITEMIDLIST)(lvitem.lParam), SHGDN_FORPARSING, &strDispName);
		StrRetToBuf(&strDispName, (LPCITEMIDLIST)(lvitem.lParam), szFileName, MAX_PATH);
		StringCchLength(szFileName, MAX_PATH, &FileNameSize);
		if (utils::File_Exists(szFileName)) {
			iFiles[iFileCounter] = iItem;
			iFileCounter++;
			for(j = 0; j < FileNameSize; j++) {
				szFullCommand[iCommandPos++] = szFileName[j];
			}
			szFullCommand[iCommandPos++] = 0;
		}
	}
	szFullCommand[iCommandPos++] = 0;
	if (iFileCounter > 0) { // In case they tried to copy my computer or something like that
		DROPFILES dobj = { 20, { 0, 0 }, 0, 1 };
		int nLen = sizeof(szFullCommand);
		int nGblLen = sizeof(dobj) + nLen*2 + 5;
		HGLOBAL hClipData = GlobalAlloc( GMEM_ZEROINIT | GMEM_MOVEABLE | GMEM_DDESHARE, nGblLen );
		char* szData = (char*)::GlobalLock(hClipData);
		memcpy( szData, &dobj, 20 );
		char* sWStr = szData+20;
		for( int i = 0; i < nLen*2; i += 2 )
		{
			sWStr[i] = szFullCommand[i/2];
		}
		::GlobalUnlock(hClipData);
		if (OpenClipboard(NULL)) {
			EmptyClipboard();
			SetClipboardData(CF_HDROP, hClipData);
			CloseClipboard();
		}
	}
}

void CGroup::PasteFiles()
{
	SHFILEOPSTRUCT SHFilePaste;
	char * buffer;
	if (!IsClipboardFormatAvailable(CF_HDROP))
		return;
	if (!OpenClipboard(NULL))
		return;
	buffer = (char*)GetClipboardData(CF_HDROP);

	char* files = buffer+20;

	char szFiles[MAX_LINE_LENGTH] = {0};

	bool bBreak = false;

	int i = 0;
	while (!bBreak) {
		szFiles[i/2] = files[i];
		if (files[i] == 0 && files[i+2]== 0) {
			bBreak = true;
		}
		i += 2;
	}

	ZeroMemory(&SHFilePaste, sizeof(SHFILEOPSTRUCT));
	SHFilePaste.wFunc = FO_COPY;
	SHFilePaste.hwnd = NULL;
	SHFilePaste.pFrom = szFiles;
	SHFilePaste.pTo = this->m_szFolderLocation;
	SHFilePaste.fFlags = 0;
	if (m_bNoConnectedFiles)
		SHFilePaste.fFlags |= FOF_NO_CONNECTED_ELEMENTS;
	SHFileOperation(&SHFilePaste);

	CloseClipboard();
}

void CGroup::RenameFile()
{
	ListView_EditLabel(m_hwndListView, ListView_GetNextItem(m_hwndListView, -1, LVNI_FOCUSED));
}

void CGroup::SortByName()
{
	ListView_SortItems(m_hwndListView, SortByNameCompareFunc, m_pFolder);
}

void CGroup::SortBySize()
{
	ListView_SortItems(m_hwndListView, SortBySizeCompareFunc, m_pFolder);
}

void CGroup::SortByType()
{
	ListView_SortItems(m_hwndListView, SortByTypeCompareFunc, m_pFolder);
}

void CGroup::SortByDate()
{
	ListView_SortItems(m_hwndListView, SortByDateCompareFunc, m_pFolder);
}

void CGroup::SetTextColor(COLORREF crColor)
{
	ListView_SetTextColor(m_hwndListView, crColor);
	// force update
	InvalidateRect(m_hwndListView, NULL, TRUE);
	UpdateWindow(m_hwndListView);
}

void CGroup::SetTextBkColor(COLORREF crColor)
{
	ListView_SetTextBkColor(m_hwndListView, crColor);
	// force update
	InvalidateRect(m_hwndListView, NULL, TRUE);
	UpdateWindow(m_hwndListView);
}

void CGroup::SetBackgroundColor(COLORREF crColor)
{
	ListView_SetBkColor(m_hwndListView, crColor);
	// force update
	InvalidateRect(m_hwndListView, NULL, TRUE);
	UpdateWindow(m_hwndListView);
}

void CGroup::SetBackground(const char *szBackground, int nOffsetPercentX, int nOffsetPercentY, bool bTiled)
{
	LVBKIMAGE lvbki;
	lvbki.ulFlags = LVBKIF_SOURCE_URL | (bTiled ? LVBKIF_STYLE_TILE : LVBKIF_STYLE_NORMAL);
	lvbki.hbm = NULL;
	lvbki.pszImage = const_cast<LPSTR>(szBackground);
	lvbki.xOffsetPercent = nOffsetPercentX;
	lvbki.yOffsetPercent = nOffsetPercentY;
	ListView_SetBkImage(m_hwndListView, &lvbki);
	// force update
	InvalidateRect(m_hwndListView, NULL, TRUE);
	UpdateWindow(m_hwndListView);
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Save/restore icons' position

void CGroup::SaveState()
{
	TRACE_LOG("CGroup::SaveState");

	HKEY hKey;
	DWORD dwDisposition;
	BYTE *pBuf;
	POINT pt;
	LPITEMIDLIST pidl;
	LVITEM lvitem;
	int nCount, nSize, npidlSize;
	char szKeyName[MAX_PATH], szNum[10];

	StringCchPrintf(szKeyName, MAX_PATH, "%s%s", REGKEY_MODULE, m_szName);

	// recreate a key

	RegDeleteKey(HKEY_CURRENT_USER, szKeyName);

	if (ERROR_SUCCESS != RegCreateKeyEx(
		HKEY_CURRENT_USER,
		szKeyName,
		0, NULL,
		REG_OPTION_NON_VOLATILE,
		KEY_ALL_ACCESS, NULL, &hKey, &dwDisposition))
	{
		TRACE_LOG("Cannot save icons' state in the registry: the key is not accessible");
		return;
	}

	// save a string (x,y,pidl) for each item
	nCount = ListView_GetItemCount(m_hwndListView);
	for (int i = 0; i < nCount; i++)
	{
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = i;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);
		pidl = (LPITEMIDLIST)(lvitem.lParam);
		ListView_GetItemPosition(m_hwndListView, i, &pt);
		npidlSize = PIDL::GetSize(pidl);
		nSize = 2 * sizeof(ULONG) + npidlSize;
		pBuf = reinterpret_cast<BYTE*>(HeapAlloc(GetProcessHeap(),0, nSize));
		((ULONG*)pBuf)[0] = pt.x;
		((ULONG*)pBuf)[1] = pt.y;
		CopyMemory(pBuf + 2 * sizeof(ULONG), pidl, npidlSize);
		StringCchPrintf(szNum, 10, "%d", i);
		RegSetValueEx(hKey, szNum, 0, REG_BINARY, pBuf, nSize);
		HeapFree(GetProcessHeap(), 0, pBuf);
	}

	RegCloseKey(hKey);
}


void CGroup::RestoreState()
{
	HKEY hKey;
	BYTE *pBuf;
	POINT pt;
	LPITEMIDLIST pidl;
	LVITEM lvitem;
	int nCount;
	char szKeyName[MAX_PATH], szNum[10];
	DWORD dwSize;

	// open the registry key

	StringCchPrintf(szKeyName, MAX_PATH, "%s%s", REGKEY_MODULE, m_szName);

	if (ERROR_SUCCESS != RegOpenKeyEx(
		HKEY_CURRENT_USER,
		szKeyName,
		0,
		KEY_ALL_ACCESS,
		&hKey))
	{
		TRACE_LOG("Cannot load icons' state from the registry: the key is not accessible");
		return;
	}

	// don't snap icons to grid
	ListView_SetExtendedListViewStyle(m_hwndListView, 0);

	// for each record in the registry key...
	for (int k = 0; ; k++)
	{
		StringCchPrintf(szNum, 10, "%d", k);
		if (RegQueryValueEx(hKey, szNum, NULL, NULL, NULL, &dwSize) != ERROR_SUCCESS)
		{
			break;
		}

		pBuf = reinterpret_cast<BYTE*>(HeapAlloc(GetProcessHeap(),0, dwSize));

		if (RegQueryValueEx(hKey, szNum, NULL, NULL, pBuf, &dwSize) != ERROR_SUCCESS)
		{
			TRACE_LOG("Damn, there is no access to the registry even for read operation!");
			break;
		}

		pt.x = ((ULONG*)pBuf)[0];
		pt.y = ((ULONG*)pBuf)[1];

		pidl = (LPITEMIDLIST)(pBuf + 2 * sizeof(ULONG));

		// try to find this pidl in the current FolderView

		nCount = ListView_GetItemCount(m_hwndListView);
		for (int i = 0; i < nCount; i++)
		{
			lvitem.mask = LVIF_PARAM;
			lvitem.iItem = i;
			lvitem.iSubItem = 0;
			ListView_GetItem(m_hwndListView, &lvitem);

			if (lvitem.lParam && m_pFolder->CompareIDs(0, pidl, (LPITEMIDLIST)(lvitem.lParam)) == 0)
			{
				ListView_SetItemPosition(m_hwndListView, i, pt.x, pt.y);
				break;
			}
		}
		HeapFree(GetProcessHeap(), 0, pBuf);
	}

	RegCloseKey(hKey);

	ListView_SetExtendedListViewStyle(m_hwndListView, m_dListStyleFlags);
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Drag&Drop support


#define HIDA_GetPIDLParent(pida) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[0])
#define HIDA_GetPIDLItem(pida, i) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[i+1])


IDropTarget* CGroup::GetLastItemTarget(bool *pbExecute)
{
	IDropTarget* rv = NULL;
	*pbExecute = false;
	if (m_nLastItem != -1)
	{
		LVITEM lvitem;
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = m_nLastItem;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);

		if (lvitem.lParam)
		{
			LPITEMIDLIST pidl = (LPITEMIDLIST)(lvitem.lParam);
			SFGAOF flags = SFGAO_DROPTARGET;
			m_pFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidl, &flags);
			bool bIsDropTarget = (flags & SFGAO_DROPTARGET) != 0;
			flags = SFGAO_FILESYSTEM;
			m_pFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidl, &flags);
			bool bIsFileSystemObject = 0 != (flags & SFGAO_FILESYSTEM);
			flags = SFGAO_FOLDER;
			m_pFolder->GetAttributesOf(1, (LPCITEMIDLIST*)&pidl, &flags);
			bool bIsFolder = 0 != (flags & SFGAO_FOLDER);
			// file system objects except folders should be managed another way...
			if (bIsFileSystemObject && !bIsFolder)
			{
				*pbExecute = true;
			}
			else if (bIsDropTarget)
			{
				m_pFolder->GetUIObjectOf(m_hWnd, 1, (LPCITEMIDLIST*)&pidl, IID_IDropTarget, NULL, (void**)&rv);
			}
		}
	}
	return rv;
}


void CGroup::Drop(DWORD grfKeyState, POINTL *ptl, DWORD *pdwEffect, CIDA *pida, POINT *pOffsets)
{
	TRACE_LOG("CGroup::Drop");

	LPITEMIDLIST pidl;	// single file object
	UINT i = 0;

	// file dropped on the icon

	if (m_nLastItem != -1)
	{
		LVITEM lvitem;
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = m_nLastItem;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);

		bool bMoveOnly = false;
		if (lvitem.lParam)
		{
			pidl = (LPITEMIDLIST)(lvitem.lParam);

			for (UINT j = 0; j < pida->cidl; j++)
			{
				if (m_pFolder->CompareIDs(0, pidl, HIDA_GetPIDLItem(pida, j)) == 0)
				{
					bMoveOnly = true;
					break;
				}
			}

			if (!bMoveOnly)
			{
				char szFileName[MAX_PATH] = {0};
				LPITEMIDLIST pidlFull = PIDL::Append(m_pidlFolder, pidl);
				if (SHGetPathFromIDList(pidlFull, szFileName))
				{
					TRACE_LOG(szFileName);
					CoTaskMemFree(pidlFull);
					while (i < pida->cidl)
					{
						char szParameter[MAX_PATH] = {0};
						pidlFull = PIDL::Append(
							HIDA_GetPIDLParent(pida),
							HIDA_GetPIDLItem(pida, i));
						if (SHGetPathFromIDList(pidlFull, szParameter))
						{
							TRACE_LOG(szParameter);
							PathQuoteSpaces(szParameter);
							ShellExecute(m_hWnd, NULL, szFileName, szParameter, NULL, SW_SHOWNORMAL);
						}
						CoTaskMemFree(pidlFull);
						++i;
					}
				}
			}
		}

		ListView_SetItemState(m_hwndView, -1, 0, LVIS_DROPHILITED);
		if (!bMoveOnly)
		{
			*pdwEffect = DROPEFFECT_COPY;
			return;
		}
	}

	while (i < pida->cidl)
	{
		// we need just to reorder icons...
		int nCount = ListView_GetItemCount(m_hwndListView);
		for (int j = 0; j < nCount; j++)
		{
			LVITEM lvitem;
			lvitem.mask = LVIF_PARAM;
			lvitem.iItem = j;
			lvitem.iSubItem = 0;
			ListView_GetItem(m_hwndListView, &lvitem);
			if (lvitem.lParam && m_pFolder->CompareIDs(0, HIDA_GetPIDLItem(pida, i), (LPITEMIDLIST)lvitem.lParam) == 0)
			{
				RECT rect;
				GetWindowRect(m_hwndListView, &rect);
				ListView_SetItemPosition(
					m_hwndListView, j,
					ptl->x + pOffsets[1 + i].x - rect.left,
					ptl->y + pOffsets[1 + i].y - rect.top);
				//m_pView2->SelectAndPositionItem(HIDA_GetPIDLItem(pida, i), SVSI_SELECT | SVSI_TRANSLATEPT, &pt);
				break;
			}
		}
		++i;
	}
	ClearDropHover();
}


void CGroup::SetDropHover(POINTL *pt, DWORD *pdwEffect)
{
	TRACE_LOG("CGroup::SetDropHover");

	// Damn, FindItem doesn't work =(

	RECT rect;

	GetWindowRect(m_hwndListView, &rect);
	pt->x -= rect.left;
	pt->y -= rect.top;

	if (m_nLastItem != -1)
	{
		ListView_GetItemRect(m_hwndListView, m_nLastItem, &rect, LVIR_BOUNDS);
		if (!(rect.left <= pt->x && pt->x < rect.right && rect.top <= pt->y && pt->y <= rect.bottom))
		{
			ListView_SetItemState(m_hwndListView, m_nLastItem, 0, LVIS_DROPHILITED);
			m_nLastItem = -1;
		}
	}

	if (m_nLastItem == -1)
	{
		int nItems = ListView_GetItemCount(m_hwndListView);
		//utils::log("(%d %d)", pt->x, pt->y);
		for (int i = 0; i < nItems; i++)
		{
			ListView_GetItemRect(m_hwndListView, i, &rect, LVIR_BOUNDS);
			//utils::log("%d (%d %d)-(%d %d)", i, rect.left, rect.top, rect.right, rect.bottom);
			if (rect.left <= pt->x && pt->x < rect.right && rect.top <= pt->y && pt->y <= rect.bottom)
			{
				m_nLastItem = i;
				ListView_SetItemState(m_hwndListView, m_nLastItem, LVIS_DROPHILITED, LVIS_DROPHILITED);
				break;
			}
		}
	}
	
	if (m_nLastItem != -1)
	{
		LVITEM lvitem;
		lvitem.mask = LVIF_PARAM;
		lvitem.iItem = m_nLastItem;
		lvitem.iSubItem = 0;
		ListView_GetItem(m_hwndListView, &lvitem);

		if (lvitem.lParam)
		{
			SFGAOF flags = SFGAO_DROPTARGET;
			m_pFolder->GetAttributesOf(1, (LPCITEMIDLIST*)(&(lvitem.lParam)), &flags);
			if ((flags & SFGAO_DROPTARGET) != SFGAO_DROPTARGET)
			{
				*pdwEffect = DROPEFFECT_NONE;
			}
		}
	}
}


void CGroup::ClearDropHover()
{
	TRACE_LOG("CGroup::ClearDropHover");
	if (m_nLastItem != -1)
	{
		ListView_SetItemState(m_hwndListView, m_nLastItem, 0, LVIS_DROPHILITED);
	}
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
// Folder View

bool CGroup::InitFolderView()
{
	m_pFolder = PIDL::GetFolder(m_pidlFolder);
	if (m_pFolder == NULL)
	{
		TRACE_LOG("No IShellFolder interface found");
		return false;
	}

	m_pFolder->CreateViewObject(m_hWnd, IID_IShellView, (void**)&m_pView);
	if (!m_pView)
	{
		TRACE_LOG("No IShellView interface found");
		return false;
	}

	FOLDERSETTINGS fs;
	fs.ViewMode = FVM_ICON; // FVM_ICON, FVM_SMALLICON, FVM_LIST, FVM_DETAILS, FVM_THUMBNAIL, FVM_TILE, FVM_THUMBSTRIP
	fs.fFlags = FWF_NOCLIENTEDGE | FWF_SNAPTOGRID /*| FWF_TRANSPARENT | FWF_DESKTOP | FWF_BESTFITWINDOW*/;
	if (!m_szBackground || !(m_szBackground[0]))
	{
		fs.fFlags |= FWF_TRANSPARENT /*| FWF_NOWEBVIEW*/ | FWF_DESKTOP;
	}
	else
	{
		fs.fFlags |= FWF_NOWEBVIEW;
	}
	if (m_bNoScroll)
	{
		fs.fFlags |= FWF_NOSCROLL;
	}
	RECT rect = { m_nX, m_nY, m_nX + m_nWidth, m_nY + m_nHeight };

	m_pView->CreateViewWindow(NULL, &fs, (IShellBrowser*)this, &rect, &m_hwndView);
	m_pView->UIActivate(SVUIA_ACTIVATE_NOFOCUS);
	if(!m_hwndView)
	{
		TRACE_LOG("Failed to create view window");
		return false;
	}
	SetWindowPos(m_hwndView, HWND_BOTTOM, 0,0,0,0, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE);

	m_pView->QueryInterface(IID_IShellView2, (void**)&m_pView2);
	if (m_pView2 == NULL)
	{
		return false;
	}

	m_hwndListView = ::GetWindow(m_hwndView, GW_CHILD);

	m_pDropTarget = new CDropTarget();
	m_pDropTarget->Register(this);

	if (m_hFont)
	{
		SendMessage(m_hwndListView, WM_SETFONT, (WPARAM)m_hFont, MAKELPARAM(TRUE, 0));
	}
	ListView_SetTextBkColor(m_hwndListView,m_crTextBackgroundColor);
	ListView_SetTextColor(m_hwndListView, m_crTextColor);
	ListView_SetIconSpacing(m_hwndListView, m_nSpacingX, m_nSpacingY);
	ListView_SetBkColor(m_hwndListView, m_crBackgroundColor);
	
	if (m_szBackground && m_szBackground[0])
	{
		SetBackground(m_szBackground, m_nOffsetPercentX, m_nOffsetPercentY, m_bTiled);
	}

	RestoreState();
	ShowWindow(m_hwndView, m_bHidden ? SW_HIDE : SW_SHOW);

	return true;
}

void CGroup::activateWndProc()
{
	if ( m_hwndListView )
	{
		SetWindowLong(m_hwndListView, GWL_USERDATA, (LONG)this);
		wpOrigListView_Proc = (WNDPROC) SetWindowLong(m_hwndListView, GWL_WNDPROC, (LONG) ListView_Proc);
	}
}

void CGroup::relayMouseMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	POINT pt;
	pt.x = GET_X_LPARAM(lParam);
	pt.y = GET_Y_LPARAM(lParam);

	MapWindowPoints(m_hwndListView, m_hWnd, &pt, 1);
	PostMessage(m_hWnd, message, wParam, MAKELPARAM((short) pt.x, (short) pt.y));
}
